% res = LevelDependentFluidStationaryDistr (masses, iniF, KF, cloF, iniB, KB, cloB, what, points)
% calculates the startionary distribution of first order and second order
% level dependent fluid models
%
% * masses: list of masses (K+1 vectors, if there are K levels)
% * iniF, KF, cloF: initial vector, matrix exponent and closing vector for
%   the forward direction for each level, a list of size K
% * iniB, KB, cloB: initial vector, matrix exponent and closing vector for
%   the backward direction for each level, a list of size K
% * what: string, can be 'pdf', 'cdf' and 'cdfm'
%     - 'pdf'  returns the (state-dependent) density for each point
%     - 'pdfd' returns the derivative of the density
%     - 'cdf'  returns the distribution function, P(X<p)
%     - 'cdfm' returns the distribution function, P(X<=p)
%
function res = LevelDependentFluidStationaryDistr (masses, iniF, KF, cloF, iniB, KB, cloB, T, what, points)

    K = length(T);
    T = [0 T];
    N = length(masses{1});

    cummPr = cell(1,K);
    for k=1:K
        cummPr{k} = expmatv2int(KF{k}, iniF{k}, T(k+1)-T(k)) * cloF{k} ...
            + expmatv2int(KB{k}, iniB{k}, T(k+1)-T(k)) * cloB{k};
    end   
   
    cummulate = strcmp(what,'cdf') || strcmp(what,'cdfm');
    res = [];
    for p=reshape(points,1,[])
        fprintf("computing cdf point %g\n", p);
        pres = zeros(1,N);
        k=0;
        while k<K && p>=T(k+1)
            if cummulate
                if k>0
                    pres = pres + cummPr{k};
                end
                if p>T(k+1) || strcmp(what,'cdfm')
                    pres = pres + masses{k+1};
                end
            end
            k = k + 1;
        end
        if k==K && p==T(k+1) && strcmp(what,'cdfm')
            pres = pres + masses{k+1};       
        end
        prem = p - T(k);
        Tk = T(k+1)-T(k);
        if strcmp(what,'pdf')
            pres = expmatv2exp(KF{k}, iniF{k}, prem) * cloF{k} + ...
                expmatv2exp(KB{k}, iniB{k}, Tk-prem) * cloB{k};
        elseif strcmp(what,'pdfd')
            pres = expmatv2exp(KF{k}, iniF{k}*KF{k}, prem) * cloF{k} - ...
                expmatv2exp(KB{k}, iniB{k}*KB{k}, Tk-prem) * cloB{k};
        elseif strcmp(what,'cdf') || strcmp(what,'cdfm')
            pres = pres + expmatv2int(KF{k}, iniF{k}, prem) * cloF{k} + ...
                expmatv2int(KB{k}, expmatv2exp(KB{k}, iniB{k}, Tk-prem), prem) * cloB{k};
        end
        res = [res; pres];
    end
end

